package edu.bu.LanguageCorrection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class AnomalyDetector {

    private static final Set<String> commonWords = new HashSet<>();
    private static final int AVERAGE_WORD_LENGTH = 5; // TODO: Change later

    static { // simple list of common words for demonstration purposes
        commonWords.add("the");
        commonWords.add("be");
        commonWords.add("to");
        commonWords.add("of");
        commonWords.add("and"); // we can add more later
    }

    public Map<String, Integer> analyzeSentences(List<String> sentences) {
        Map<String, Integer> sentenceScores = new HashMap<>();
        for (String sentence : sentences) {
            int score = 0;
            // length variance
            score += Math.abs(sentence.length() - AVERAGE_WORD_LENGTH * 10); // Assuming an average sentence length

            // word rarity
            String[] words = sentence.split("\\s+");
            for (String word : words) {
                if (!commonWords.contains(word.toLowerCase())) {
                    score += 10; // Increment score for each UNCOMMON word
                }
            }

            sentenceScores.put(sentence, Math.min(score, 100)); // Limit to 100 (normalize)
        }
        return sentenceScores;
    }

    public Map<String, Integer> analyzePhrases(List<String> sentences) {
        Map<String, Integer> phraseScores = new HashMap<>();

        for (String sentence : sentences) {
            List<String> phrases = TextProcessor.extractPhrases(sentence, 2, 3);

            for (String phrase : phrases) {
                int score = 0;
                // Score length variance like above
                if (phrase.length() < AVERAGE_WORD_LENGTH || phrase.length() > AVERAGE_WORD_LENGTH * 3) {
                    score += 20;
                }
                // Word rarity
                String[] words = phrase.split("\\s+");
                for (String word : words) {
                    if (!commonWords.contains(word.toLowerCase())) {
                        score += 5;
                    }
                }
                phraseScores.put(phrase, Math.min(score, 100)); // Normalize
            }
        }

        return phraseScores;
    }
}
